import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;
import R2Graphics.*;

/**
 * The R2Convex class computes a convex of a point set on a 2-dimensional plane
 */
public class R2Convex extends Object {
    R2Point         a;          // First vertex of the line segment
    R2Point         b;          // Second vertex of line segment
    int             numAngles;  // Number of verteces: 0, 1, 2, many
    ConvexPolygon   polygon;    // Pointer to polygon

    public R2Convex() {
        a = null; b = null; numAngles = 0; polygon = null;
    }

    public void init() {
        a = null; b = null; numAngles = 0; polygon = null;
    }

    public double area() {
        if (numAngles < 3) {
            return 0.;
        } else {
            return polygon.area();
        }
    }

    public double perimeter() {
        if (numAngles < 2) {
            return 0.;
        } else if (numAngles == 2) {
            return 2. * a.distance(b);
        } else {
            return polygon.perimeter();
        }
    }
	
//
//------------------------------------------------------
//Finds the maximum of all angles between our point and all
//points of the polygon
//
    public double angle(R2Point p) {
	Enumeration points = elements();
	R2Point nextPoint = new R2Point(0., 0.);
	R2Point nextPoint2 = new R2Point(0., 0.);
	double maxAngle = -1.;
	double nextAngle;

	while(points.hasMoreElements()) {
	    nextPoint = (R2Point)points.nextElement();
	    Enumeration points2 = elements();
	    while(points2.hasMoreElements()) {
		nextPoint2 = (R2Point)points2.nextElement();
		nextAngle = calcAngle(nextPoint, nextPoint2, p);
		if(nextAngle > maxAngle) {
		    maxAngle = nextAngle;
		}
	    }
	}

	return maxAngle;
    }

//Calculate the angle with the use of cosinus theorem
    private double calcAngle(R2Point x1, R2Point x2, R2Point a) {
	double cosAngle;
	double radAngle;
	double gradAngle;
	String result;
	
        cosAngle = (a.distance(x1)*a.distance(x1)+a.distance(x2)*a.distance(x2)-x1.distance(x2)*x1.distance(x2))/(2*a.distance(x1)*a.distance(x2));
	radAngle = Math.acos(cosAngle);
	gradAngle = 180*radAngle/Math.PI;
	
	return gradAngle;
    }
//------------------------------------------------------
//

    public void add(R2Point t) {
        if (numAngles == 0) {
            numAngles++;
            a = t;
        } else if (numAngles == 1) {
            if (!t.equals(a)) {
                numAngles++;
                b = t;
            }
        } else if (numAngles == 2) {
            if (!t.onLine(a, b)) {
                polygon = new ConvexPolygon(a, b, t);
                numAngles++;
            } else if (a.between(t, b)) {
                a = t;
            } else if (b.between(a, t)) {
                // b between a and t
                b = t;
            }
        } else {
            // Pass the request to polygon
            polygon.add(t);
        }
    }

    // Enumeration of convex elements
    class R2ConvEnum extends Object implements Enumeration {
        private int curPoint = 0;

        public boolean hasMoreElements() {
            return curPoint < numAngles;
        }

        public Object nextElement() throws NoSuchElementException {
            R2Point p;
            if (curPoint < numAngles) {
                if (curPoint == 0) p = a;
                else               p = b;
                curPoint++;
                return p;
            } else {
                throw new NoSuchElementException(
                    "No more points in the convex"
                );
            }
        }
    }

    public final synchronized Enumeration elements() {
        if (numAngles <= 2) {
            return new R2ConvEnum();
        } else {
            return polygon.elements();
        }
    }
}

class ConvexPolygon extends Object {
    double      polygonArea;
    double      polygonPerimeter;
    Deq  deq;

    public ConvexPolygon(R2Point a, R2Point b, R2Point c) {
        // Put point in deq in order (p1, p2, p3) such that
        // p3-->p1 is not lit from p2. It implies that
        // p1-->p2 is not lit from p3 and p2-->p3 is not lit from p1,
        // i.e. the triangle edges are not lit from triangle inner points.
        deq = new Deq();
        deq.pushBeg(b);
        if (lit(a, c, b)) {
            deq.pushBeg(a);
            deq.pushEnd(c);
        } else {
            deq.pushBeg(c);
            deq.pushEnd(a);
        }

        polygonPerimeter = a.distance(b) + b.distance(c) + c.distance(a);
        polygonArea = a.area(b, c); // Triangle area
    }

    public double area() {
        return polygonArea;
    }

    public double perimeter() {
        return polygonPerimeter;
    }


    public void add(R2Point t) {
        int i; R2Point x;

        // Look up the lit edge
        for (i = 0; i < deq.size(); i++) {
            if (lit((R2Point) deq.end(), (R2Point) deq.beginning(), t))
                break;
            // Rotate the polygon
            deq.pushEnd(
                deq.popBeg()
            );
        }

        // If current edge is not lit, then nothing to do :-)
        if (!lit((R2Point) deq.end(), (R2Point) deq.beginning(), t))
            return;

        // Assertion: current edge is lit from the point t.

        // Delete the current edge
        // (i.e. modify polygonArea and polygonPerimeter)
        deleteEdge((R2Point) deq.end(), (R2Point) deq.beginning(), t);

        // Delete lit edges from the beginning of deq
        x = (R2Point) deq.popBeg();
        while (lit(x, (R2Point) deq.beginning(), t)) {
            deleteEdge(x, (R2Point) deq.beginning(), t);
            x = (R2Point) deq.popBeg();
        }
        deq.pushBeg(x);

        // Delete lit edges from the end of deq
        x = (R2Point) deq.popEnd();
        while (lit((R2Point) deq.end(), x, t)) {
            deleteEdge((R2Point) deq.end(), x, t);
            x = (R2Point) deq.popEnd();
        }
        deq.pushEnd(x);

        polygonPerimeter += t.distance((R2Point) deq.beginning()) +
            t.distance((R2Point) deq.end());

        deq.pushBeg(t);
    }

    /**
     * The edge a-->b is lit from a point t.
     * Returns true, if a triangle <a, b, t> is
     * counterclockwise oriented.
     */
//
//--------------------------------------------------------
//You need to change this method from private to public!!!
//--------------------------------------------------------
//
    public static boolean lit(R2Point a, R2Point b, R2Point t) {
        double triangArea = t.signedArea(a, b);
        return (
            triangArea < (-R2Point.PRESIZION) ||
            (
                Math.abs(triangArea) <= R2Point.PRESIZION &&
                ! t.between(a, b)
            )
        );
    }

    /**
     * Modify the area and perimeter after deleting of the edge a-->b,
     * when adding a new vertex t.
     */
    private void deleteEdge(R2Point a, R2Point b, R2Point t) {
        polygonPerimeter -= a.distance(b);
        polygonArea += t.area(a, b);
    }

    /**
     * Enumeration of polygon vertices
     */
    public final synchronized Enumeration elements() {
        return deq.elements();
    }
}

/**
 * DEQ (Double-Ended Queue)
 */
class Deq extends Object {
    static final int DEQ_INITIAL_SIZE = 1024;
    int    numElements = 0;
    int    beg = 0;             // Index of beginning
    int    end = (-1);          // Index of end
    Vector elementsArray = null;
    int    maxSize = 0;

    public Deq() {
        maxSize = DEQ_INITIAL_SIZE;
        elementsArray = new Vector(maxSize); // Set initial max. size of Deq
        elementsArray.setSize(maxSize);
        removeAllElements();
    }

    public Deq(int initialMaxSize) {
        maxSize = initialMaxSize;
        elementsArray = new Vector(maxSize); // Set initial max. size of Deq
        elementsArray.setSize(maxSize);
        removeAllElements();
    }

    public void removeAllElements() {
        numElements = 0;
        beg = 0; end = maxSize - 1;
    }

    public boolean empty() { return numElements == 0; }

    public int size() { return numElements; }

    private int nextIndex(int i) {
        if (i < maxSize - 1) return (i + 1);
        else return 0;
    }

    private int prevIndex(int i) {
        if (i > 0) return (i - 1);
        else return maxSize - 1;
    }

    private void extendDeq() { // Make maximal size of deq larger
        elementsArray.setSize(2 * maxSize);
        for (int i = 0; i < beg; i++) {
            elementsArray.setElementAt(
                elementsArray.elementAt(i),
                maxSize + i
            );
        }
        end = beg + maxSize - 1;
        maxSize *= 2;
    }

    public void pushBeg(Object p) {
        if (numElements >= maxSize)
            extendDeq();
        beg = prevIndex(beg);
        elementsArray.setElementAt(p, beg);
        numElements++;
    }

    public void pushEnd(Object p) {
        if (numElements >= maxSize)
            extendDeq();
        end = nextIndex(end);
        elementsArray.setElementAt(p, end);
        numElements++;
    }

    public Object popBeg() throws DeqException {
        if (empty()) {
            throw new DeqException("Deq is empty");
        }
        Object p = elementsArray.elementAt(beg);
        beg = nextIndex(beg);
        numElements--;
        return p;
    }

    public Object popEnd()  throws DeqException {
        if (empty()) {
            throw new DeqException("Deq is empty");
        }
        Object p = elementsArray.elementAt(end);
        end = prevIndex(end);
        numElements--;
        return p;
    }

    public Object beginning() throws DeqException {
        if (empty()) {
            throw new DeqException("Deq is empty");
        }
        return elementsArray.elementAt(beg);
    }

    public Object end() throws DeqException {
        if (empty()) {
            throw new DeqException("Deq is empty");
        }
        return elementsArray.elementAt(end);
    }

    /*
     * Enumeration of DEQ elements
     */
    class DeqEnum extends Object implements Enumeration {
        int curPoint;
        int elementNumber;

        public DeqEnum() { curPoint = beg; elementNumber = 0; }

        public boolean hasMoreElements() {
            return (elementNumber < numElements);
        }

        public Object nextElement() throws NoSuchElementException {
            if (curPoint != nextIndex(end)) {
                Object p = elementsArray.elementAt(curPoint);
                curPoint = nextIndex(curPoint);
                elementNumber++;
                return p;
            } else {
                throw new NoSuchElementException(
                    "No more points in the DEQ"
                );
            }
        }
    }

    public final synchronized Enumeration elements() {
        return new DeqEnum();
    }
}

class DeqException extends RuntimeException {
    public DeqException() {}
    public DeqException(String reason) {
        super(reason);
    }
}
